# GNU GPL LICENSE
#
# This module is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; latest version thereof,
# available at: <http://www.gnu.org/licenses/gpl.txt>.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this module; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

"""
Provides a :func:`collection.namedtuple` for :class:`Accession` and a
:class:`dict` to hold the :class:`Mapping`.
"""

from collections import defaultdict, namedtuple

__author__  = "Florian Leitner"
__version__ = "1.0"

Accession = namedtuple('Accession', 'acc tax')
"""A :obj:`namedtuple` with the attributes :attr:`acc` (:obj:`str`) and
:attr:`tax` (:obj:`int`).
"""

class Mapping(dict):
    """
    Mapping data structure::
      
      { src_acc: { tax_id: [ trgt_acc, trgt_acc, ... ], ... }, ... }
    """
    
    def set_source(self, source):
        """
        Set the source accession for the :meth:`add_target` method.
        
        :arg source: source accession
        :type source: :class:`str`
        """
        if source not in self:
            self[source] = defaultdict(list)
        
        self._src = self[source]
    
    def add_target(self, target):
        """
        Add a new target :class:`Accession` for the currently
        :meth:`set_source`.
        
        :arg target: to add
        :type target: :class:`Accession`
        """
        self._src[target.tax].append(target.acc)
    
    def targets(self):
        """
        Get a list of all target dictionaries.
        
        :returns: targets (as ``{ int(tax_id): [ str(acc), ... ], ... }``)
        :rtype: :class:`list`
        """
        return self.values()
    
    def sources(self):
        """
        Get a list of all source accessions.
        
        :returns: source accessions (:obj:`str`)
        :rtype: :class:`list`
        """
        return self.keys()
    
    def taxa(self):
        """
        Get a (unique) list of all target taxIDs.
        
        :return: all tax IDs (:obj:`int`) to which a mapping has been made
        :rtype: :class:`list`
        """
        target_taxa = reduce(lambda a, b: a|b, self._itertaxa())
        return list(target_taxa)
    
    def _itertaxa(self):
        "Yield target taxa key sets."
        for target in self.itervalues():
            yield set(target.keys())
    

