# GNU GPL LICENSE
#
# This module is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; latest version thereof,
# available at: <http://www.gnu.org/licenses/gpl.txt>.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this module; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

"""
Functions :func:`hhmap.extract` uses to create HH mappings:
first :func:`find_mappings`, then :func:`print_mappings`.
"""

import logging
from cStringIO import StringIO
from .data import Accession, Mapping

__author__  = "Florian Leitner"
__version__ = "1.0"

def extract(sources, clusters, taxa=None, excluding=True, ungrouped=False):
    """
    Parses *clusters* by searching for all mappings for the *sources*
    (limited to a set of *taxa* and/or by *excluding* the taxon of the
    source accession) and prints the found mappings per source on one line.
    
    The found (target) accessions are grouped by taxIDs or can be
    printed *ungrouped* (all together in on CSV list). Maps to all found
    accessions if *taxa* is an empty list or :obj:`None`, and prints
    *ungrouped* if *taxa* is :obj:`None`.
    
    The *clusters* should be an iterator over each cluster, returning
    another iterator for the ``taccession,taxID`` strings.
    
    :arg  sources: lookup accessions (:obj:`str`) for mappings 
    :type sources: :class:`list`
    
    :arg  clusters: a cluster iterator
    :type clusters: :func:`iter`
    
    :arg  taxa: taxIDs (:obj:`int`) to map to
    :type taxa: :class:`list` or :obj:`None`
    
    :arg  excluding: exclude mappings to the same taxon as the source
    :type excluding: :class:`bool`
    
    :arg  ungrouped: do not split target accessions by taxa in output
    :type ungrouped: :class:`bool`
    """
    targets = None
    
    if taxa:
        targets = set(taxa)
        logging.debug("target taxa %s", targets)
    
    logging.info("parsing mappings")
    mappings = find_mappings(sources, clusters, targets, excluding)
    
    if ungrouped or taxa is None:
        taxa = None
    else:
        logging.info("collecting found taxa")
        taxa = mappings.taxa()
    
    logging.info("printing mappings")
    print_mappings(mappings, taxa)
    
    return 0

def find_mappings(accessions, clusters, taxa=None, exclude=False):
    """
    Extract mappings (limited by *taxa*) from *clusters* for the given
    source *accessions*.
    
    If *taxa* is ``None``, all possible mappings are retrieved. If *exclude*
    is ``True``, mappings to accessions with the same taxID as the source
    are skipped.
    
    :arg  accessions: lookup (source) accessions (:obj:`str`) for mappings 
    :type accessions: :class:`list`
    
    :arg  clusters: a cluster iterator - see :func:`hhmap.extract`
    :type clusters: :func:`iter`
    
    :arg  taxa: taxIDs (:obj:`int`) to map to (all if :obj:`None`)
    :type taxa: :class:`set` or :obj:`None`
    
    :arg  exclude: exclude mappings to the same taxon as the source
    :type exclude: :class:`bool`
    
    :returns: the found mappings
    :rtype:   :class:`hhmap.data.Mapping`
    """
    mappings = Mapping()
    
    for cluster in iter(clusters):
        sources = []
        targets = []
        
        for item in iter(cluster):
            acc, tax = item.split(',')
            item = Accession(acc, int(tax))
            
            if item.acc in accessions:
                logging.debug("found source %s", item)
                sources.append(item)
            elif taxa is None:
                targets.append(item)
            elif item.tax in taxa:
                targets.append(item)
        
        if len(sources) and len(targets):
            for src in sources:
                mappings.set_source(src.acc)
                
                for trgt in targets:
                    if not exclude or trgt.tax != src.tax:
                        mappings.add_target(trgt)
                    else:
                        logging.debug("excluding %s", trgt)
                
                if exclude and not len(mappings[src.acc]):
                    del mappings[src.acc]
                else:
                    logging.info("found mappings for source %s", src)
    
    return mappings

def print_mappings(mappings, taxa=None, header=True):
    """
    Print found *mappings* for the list of *taxa* as tab-separated values.
    
    The first column contains the source accession, all following columns
    contain mapped accessions for a specific tax ID per column. If *header*
    is ``False``, the header line with the tax IDs for the target taxa
    columns is omitted. If *taxa* is ``None`` or an empty list, mappings are
    printed ungrouped and the header is not printed.
    
    :arg  mappings: as generated by :func:`hhmap.mapper.find_mappings`
    :type mappings: :class:`hhmap.data.Mapping`
    
    :arg  taxa: tax IDs (:class:`int`) for which mappings should be printed
    :type taxa: :class:`list` or :obj:`None`
    
    :arg  header: print header line containing the target taxa IDs
                  ()
    :type header: :class:`bool`
    """
    str_buffer = StringIO()
    
    if taxa and header:
        str_buffer.write("source\t%s\n" % "\t".join(str(t) for t in taxa))
    
    if taxa:
        str_buffer = _mappings_by_tax_id(str_buffer, mappings, taxa)
    else:
        str_buffer = _mappings_ungrouped(str_buffer, mappings)
    
    print str_buffer.getvalue(),

def _mappings_by_tax_id(str_buffer, mappings, taxa):
    "Print target accessions to *str_buffer*, tab-separated by *taxa*."
    for s_acc, targets in mappings.iteritems():
        if any((tax_id in targets) for tax_id in taxa):
            str_buffer.write(s_acc)
            
            for tax_id in taxa:
                str_buffer.write("\t")
                
                if tax_id in targets:
                    str_buffer.write(",".join(targets[tax_id]))
                else:
                    str_buffer.write("")
            
            str_buffer.write("\n")
    
    return str_buffer

def _mappings_ungrouped(str_buffer, mappings):
    "Print all target accessions to *str_buffer* in one CSV column."
    for s_acc, targets in mappings.iteritems():
        str_buffer.write(s_acc)
        str_buffer.write("\t")
        str_buffer.write(",".join(
            ",".join(t_accs) for t_accs in targets.values()
        ))
        str_buffer.write("\n")
    
    return str_buffer

